/* DO NOT EDIT! GENERATED AUTOMATICALLY! */

/* Virtual function table layout of superclass.  */
#include "styled_ostream.vt.h"

/* Virtual function table layout of noop_styled_ostream class.  */
      ostream_t (*get_destination) (THIS_ARG);
   bool      (*is_owning_destination) (THIS_ARG);
